/**************************************************************************************
Author:			David Szybala
Program:		Spline_Tools.ms
Date:			04/18/11
Description:	This script provides a number of methods for working with splines
***************************************************************************************/

try (destroyDialog spline_tools_rollout) catch()


struct SplineTools
(
    /**************************************************************************
    Returns valid spline objects that are selected
    **************************************************************************/
    fn get_selected_splines =
    (
        local selected_splines = for current in selection \
                                    where (classOf current == SplineShape) \
                                        or (classOf current == line) collect current
        
        selected_splines --return value
    ), --end fn
    
    /**************************************************************************
    Returns the knot selection for 'spline_object' as an array of
    #(<spline index>, #(<selected knots>))
    **************************************************************************/
    fn get_knot_selection spline_object =
	(
		local selected_spline_knots = #()
		local current_knots = #()
		
		for index = 1 to (numSplines spline_object) do
		(
			current_knots = getKnotSelection spline_object index
			
			if current_knots.count > 0 do
			(
				append selected_spline_knots index
				append selected_spline_knots current_knots
			)--end if
		)--end for
		
		selected_spline_knots --return value
	), --end fn
    
    /**************************************************************************
    Sets the IN tangent to match the out tangent for 'spline_knot' on
    'spline_object' in 'spine_index'
    **************************************************************************/
    fn set_tangent_in_to_out spline_object spline_index spline_knot =
	(
		local knot_position = getKnotPoint spline_object spline_index spline_knot
		local out_tangent = getOutVec spline_object spline_index spline_knot
		
		setInVec spline_object spline_index spline_knot ((knot_position - out_tangent) + knot_position)
	), --end fn
	
	/**************************************************************************
    Sets the OUT tangent to match the in tangent for 'spline_knot' on
    'spline_object' in 'spine_index'
    **************************************************************************/
	fn set_tangent_out_to_in spline_object spline_index spline_knot =
	(
		local knot_position = getKnotPoint spline_object spline_index spline_knot
		local in_tangent = getInVec spline_object spline_index spline_knot
		
		setOutVec spline_object spline_index spline_knot ((knot_position - in_tangent) + knot_position)
	), --end fn
    
    /**************************************************************************
    Sets the tangent for all selected knots in 'spline_objects' based on the
    'mode'
    **************************************************************************/
    fn set_knot_tangents spline_objects mode =
    (
        local selected_spline_knots = #()
			
        for current in spline_objects do
        (
            selected_spline_knots = this.get_knot_selection current
            
            for i = 1 to selected_spline_knots.count by 2 do
            (
                for knot in selected_spline_knots[i + 1] do
                (
                    if mode == 1 then
                        this.set_tangent_in_to_out current selected_spline_knots[i] knot
                    else
                        this.set_tangent_out_to_in current selected_spline_knots[i] knot
                )--end for
            )--end for
        
            updateShape current
        )--end for
    ), --end fn
    
    /**************************************************************************
    Sets the knot type for all selected knots in 'spline_objects' based on the
    'mode'
    **************************************************************************/
    fn set_knot_type spline_objects mode =
    (
        local selected_spline_knots = #()
			
        for current in spline_objects do
        (
            selected_spline_knots = this.get_knot_selection current
            
            for i = 1 to selected_spline_knots.count by 2 do
            (
                for knot in selected_spline_knots[i + 1] do
                    setKnotType current selected_spline_knots[i] knot mode
            )--end for
        
            updateShape current
        )--end for
    ), --end fn
    
    /**************************************************************************
    Returns the vector of 'knot_mid' based on 'knot_start' and 'knot_end' for
    'spline_object' on 'spline_index'
    **************************************************************************/
    fn get_knot_vector spline_object spline_index knot_start knot_mid knot_end =
	(
		local vector1 = (getKnotPoint spline_object spline_index knot_start) - (getKnotPoint spline_object spline_index knot_mid)
		local vector2 = (getKnotPoint spline_object spline_index knot_mid) - (getKnotPoint spline_object spline_index knot_end)
		local vector3 = acos (dot (normalize vector1) (normalize vector2))
		
		vector3 --return value
	), --end fn
    
    /**************************************************************************
    Selects all mid edge knots for all 'spline_objects' Returns the total
    number of knots selected
    **************************************************************************/
    fn select_mid_edge_knots spline_objects angle_threshold =
    (
        local knot_vector = point3 0 0 0
        local mid_edge_knots = #()
        local counter = 0
        
        for current in spline_objects do
        (	
            for i = 1 to (numSplines current) do
            (
                setKnotSelection current i #() keep:false
                
                if isClosed current i then
                (
                    --check first vertex
                    knot_vector = this.get_knot_vector current i (numKnots current i) 1 2
                    
                    if knot_vector < angle_threshold do
                        append mid_edge_knots 1
                    
                    --check all
                    for x = 1 to ((numKnots current i) - 2) do
                    (
                        knot_vector = this.get_knot_vector current i x (x + 1) (x + 2)
                        
                        if knot_vector < angle_threshold do
                            append mid_edge_knots (x + 1)
                    )--end for
                    
                    --check last vertex
                    knot_vector = this.get_knot_vector current i ((numKnots current i) - 1) (numKnots current i) 1
                    
                    if knot_vector < angle_threshold do
                        append mid_edge_knots (numKnots current i)
                )--end if
                else
                (
                    for x = 1 to ((numKnots current i) - 2) do
                    (
                        knot_vector = this.get_knot_vector current i x (x + 1) (x + 2)
                        
                        if knot_vector < angle_threshold do
                            append mid_edge_knots (x + 1)
                    )--end for
                )--end else
                
                if mid_edge_knots.count > 0 do
                (
                    setKnotSelection current i mid_edge_knots keep:true
                    counter += mid_edge_knots.count
                ) --end if
                
                free mid_edge_knots
            )--end for
        )--end for
        
        counter --return value
    ), --end fn
    
    /**************************************************************************
    Returns true if an even number of knots have been selected on 'spline_object'
    **************************************************************************/
    fn validate_knot_count spline_object =
    (
        local selected_spline_knots = this.get_knot_selection spline_object
        local loop_control = true
        local is_valid = true
        
        for i = 1 to selected_spline_knots.count by 2 while loop_control do
        ( 
            if ((mod selected_spline_knots[i + 1].count 2) as integer) != 0 do
            (
                is_valid = false
                loop_control = false
            ) --end if
        ) --end for
        
        is_valid --return value
    ), --end fn
    
    /**************************************************************************
    Returns true if knots have been selected in pairs correctly on 'spline_object'
    **************************************************************************/
    fn validate_knot_pairs spline_object =
    (
        local selected_spline_knots = this.get_knot_selection spline_object
        local loop_control = true
        local is_valid = true
        
        for i = 1 to selected_spline_knots.count by 2 while loop_control do
        (
            selected_knots = selected_spline_knots[i + 1]
            
            for x = 1 to selected_knots.count by 2 while loop_control do
            (
                if (selected_knots[x] + 1) != selected_knots[x + 1] do
                (
                    is_valid = false
                    loop_control = false
                ) --end if
            ) --end for
        ) --end for
        
        is_valid --return value
    ), --end fn
    
    /**************************************************************************
    Sets the tangents for 'knot1' and 'knot2' to a smooth value of
    'smooth_percent'
    **************************************************************************/
    fn set_knot_smoothing spline_object spline_index knot1 knot2 smooth_percent =
	(
        local total_knots = numKnots spline_object spline_index
		local pos1, pos2, pos3, pos4
		local dist1, dist2, dist3
		
		if (knot1 - 1) > 0 then
			pos1 = getKnotPoint spline_object spline_index (knot1 - 1)
		else
			pos1 = getKnotPoint spline_object spline_index total_knots
			
		pos2 = getKnotPoint spline_object spline_index knot1
		pos3 = getKnotPoint spline_object spline_index knot2
		
		if (knot2 + 1) <= total_knots then
			pos4 = getKnotPoint spline_object spline_index (knot2 + 1)
		else
			pos4 = getKnotPoint spline_object spline_index 1
		
		dist1 = distance pos1 pos2
		dist2 = distance pos2 pos3
		dist3 = (dist2 / dist1) * (smooth_percent * 0.01)
		
		setKnotType spline_object spline_index knot1 #bezier
		setOutVec spline_object spline_index knot1 (pos2 + ((pos2 - pos1) * dist3))
		this.set_tangent_in_to_out spline_object spline_index knot1
		
		dist1 = distance pos4 pos3
		dist2 = distance pos3 pos2
		dist3 = (dist2 / dist1) * (smooth_percent * 0.01)
		
		setKnotType spline_object spline_index knot2 #bezier
		setInVec spline_object spline_index knot2 (pos3 + ((pos3 - pos4) * dist3))
		this.set_tangent_out_to_in spline_object spline_index knot2
	) --end fn    
) --end struct


(--put script into local scope
	global spline_tools_rollout
	local spline_tools_ini_file = getDir #plugcfg + "\\SplineTools.ini"
	local spline_tools_position

	rollout spline_tools_rollout "Spline Tools"
	(
        local spline_tools = SplineTools()
        
		groupBox tangent_group "Tangents" pos:[5,5] width:190 height:70
			button in_tangent_button "Make In Match Out" pos:[15,25] width:170 height:20
			button out_tangent_button "Make Out Match In" pos:[15,50] width:170 height:20
        
        groupBox knot_type_group "Convert Knots" pos:[5,80] width:190 height:70
			button smooth_knot_button "Smooth" pos:[15,100] width:80 height:20
			button corner_knot_button "Corner" pos:[105,100] width:80 height:20
			button bezier_knot_button "Bezier" pos:[15,125] width:80 height:20
			button bezier_corner_knot_button "Bezier Corner" pos:[105,125] width:80 height:20
    
        groupBox mid_edge_group "Mid Edge Knots" pos:[5,155] width:190 height:65
			label angle_label "Angle Threshold:" pos:[15,175] width:84 height:15
			spinner angle_threshold_spinner "" pos:[140,173] range:[0.01,45,0.1] width:45 height:16
			button mid_edge_button "Select Mid Edge Knots" pos:[15,195] width:170 height:20
			
		groupBox smooth_group "Smooth Tangents" pos:[5,225] width:190 height:65
			label smooth_label "Smoothness (%):" pos:[15,245] width:85 height:15
			spinner smooth_tangent_spinner "" pos:[135,243] range:[0.0,100.0,50.0] type:#float scale:1.0 width:50 height:15
			button smooth_tangent_button "Smooth Tangents Pairs" pos:[15,265] width:170 height:20
        
		groupBox misc_group "Misc" pos:[5,295] width:190 height:45
			button open_shape_button "Select Open Splines" pos:[15,315] width:170 height:20
        
        label status_label "INFO: Standing by..." pos:[5,350] width:190 height:15
		
        /**********************************************************************
        UI move event handler. Stores the curent position of the UI
        **********************************************************************/
		on spline_tools_rollout moved pos do
			setIniSetting spline_tools_ini_file "Dialog" "Position" (pos as string)
			
		/**********************************************************************
        In match out event handler. Makes the in tangents on all selected knots
        match their corresponding out tangents
        **********************************************************************/
		on in_tangent_button pressed do
		(
			local selected_splines = spline_tools.get_selected_splines()
            
            if selected_splines.count == 0 do
            (
                status_label.text = "ERROR: No splines selected"
                return false
            ) --end if
            
            spline_tools.set_knot_tangents selected_splines 1
            
            status_label.text = "INFO: In tangents match out"
        ) --end on
		
		/**********************************************************************
        Out match in event handler. Makes the out tangents on all selected knots
        match their corresponding in tangents
        **********************************************************************/
		on out_tangent_button pressed do
		(
			local selected_splines = spline_tools.get_selected_splines()
            
            if selected_splines.count == 0 do
            (
                status_label.text = "ERROR: No splines selected"
                return false
            ) --end if
            
            spline_tools.set_knot_tangents selected_splines 2
            
            status_label.text = "INFO: Out tangents match in"
		)--end on
        
        /**********************************************************************
        Set knot type smooth event handler. Sets the selected knots on all
        selected splines to smooth
        **********************************************************************/
        on smooth_knot_button pressed do
        (
            local selected_splines = spline_tools.get_selected_splines()
            
            if selected_splines.count == 0 do
            (
                status_label.text = "ERROR: No splines selected"
                return false
            ) --end if
            
            spline_tools.set_knot_type selected_splines #smooth
            
            status_label.text = "INFO: Knots set to smooth"
        ) --end on
        
        /**********************************************************************
        Set knot type corner event handler. Sets the selected knots on all
        selected splines to corner
        **********************************************************************/
        on corner_knot_button pressed do
        (
            local selected_splines = spline_tools.get_selected_splines()
            
            if selected_splines.count == 0 do
            (
                status_label.text = "ERROR: No splines selected"
                return false
            ) --end if
            
            spline_tools.set_knot_type selected_splines #corner
            
            status_label.text = "INFO: Knots set to corner"
        ) --end on
        
        /**********************************************************************
        Set knot type bezier event handler. Sets the selected knots on all
        selected splines to bezier
        **********************************************************************/
        on bezier_knot_button pressed do
        (
            local selected_splines = spline_tools.get_selected_splines()
            
            if selected_splines.count == 0 do
            (
                status_label.text = "ERROR: No splines selected"
                return false
            ) --end if
            
            spline_tools.set_knot_type selected_splines #bezier
            
            status_label.text = "INFO: Knots set to bezier"
        ) --end on
        
        /**********************************************************************
        Set knot type bezier corner event handler. Sets the selected knots on all
        selected splines to bezier corner
        **********************************************************************/
        on bezier_corner_knot_button pressed do
        (
            local selected_splines = spline_tools.get_selected_splines()
            
            if selected_splines.count == 0 do
            (
                status_label.text = "ERROR: No splines selected"
                return false
            ) --end if
            
            spline_tools.set_knot_type selected_splines #bezierCorner
            
            status_label.text = "INFO: Knots set to bezier corner"
        ) --end on
        
        /**********************************************************************
        Mid edge knot event handler. Selects all mid edge knots based on the
        threshold for all selected spines 
        **********************************************************************/
        on mid_edge_button pressed do
		(
            local selected_splines = spline_tools.get_selected_splines()
            
            if selected_splines.count == 0 do
            (
                status_label.text = "ERROR: No splines selected"
                return false
            ) --end if
            
            result = spline_tools.select_mid_edge_knots selected_splines angle_threshold_spinner.value
            
            status_label.text = ("INFO: Selected " + (result as string) + " mid edge knots")
        ) --end on
            
        /**********************************************************************
        Smooth corner event handler. Smooths the tangents between all
        selected knot pairs
        **********************************************************************/
		on smooth_tangent_button pressed do
		(
            local selected_splines = spline_tools.get_selected_splines()
            
            if selected_splines.count == 0 do
            (
                status_label.text = "ERROR: No splines selected"
                return false
            ) --end if
            
            for current in selected_splines do
            (
                --CRITICAL! Make sure an even number of knots have been selected!
                if not spline_tools.validate_knot_count current do
                (
                    print("ERROR: Spline object " + current.name + " does not have an even number of knots selected")
                    status_label.text = ("ERROR: Check listener for details")
                    return false
                ) --end if
                
                --CRITICAL! Make sure proper knot pairs have been selected!
                if not spline_tools.validate_knot_pairs current do
                (
                    print("ERROR: Spline object " + current.name + " does not have proper 'knot pairs' selected")
                    status_label.text = ("ERROR: Check listener for details")
                    return false
                ) --end if
            ) --end for
            
            for current in selected_splines do
            (
                selected_spline_knots = spline_tools.get_knot_selection current
            
                for i = 1 to selected_spline_knots.count by 2 do
                (
                    selected_knots = selected_spline_knots[i + 1]
            
                    for x = 1 to selected_knots.count by 2 do
                        spline_tools.set_knot_smoothing current i selected_knots[x] selected_knots[x + 1] smooth_tangent_spinner.value
                ) --end for
                
                updateShape current
            ) --end for
                
            status_label.text = ("INFO: Tangent smoothing complete")
        ) --end on
            
		/**********************************************************************
        Open shape event handler. Selects all open shapes
        **********************************************************************/
		on open_shape_button pressed do
		(
			local spline_objects = for current in shapes where (classOf current == SplineShape) collect current
            local open_splines = #()
			local counter = 0
            
            clearSelection()
			
			if spline_objects.count > 0 then
			(
				for current in spline_objects do
				(
					for i = 1 to (numSplines current) do
					(
						if not (isClosed current i) do
						(
                            appendIfUnique open_splines current
							print ("WARNING: " + current.name + " spline #" + (i as string) + " is not closed")
							counter += 1
						)--end if
					)--end for
				)--end for
				
				print ("WARNING: Total open splines: " + (counter as string))
			)--end if
            
            if counter == 0 then
                status_label.text = "INFO: No open splines found"
            else
            (
                select open_splines
                status_label.text = "WARNING: Check listener for details"
            ) --end else
		)--end on
	)--end rollout
    
	spline_tools_position = execute (getIniSetting spline_tools_ini_file "Dialog" "Position")
	if spline_tools_position == OK do spline_tools_position = [100,100]
	if spline_tools_position.x > sysinfo.desktopsize.x - 200 do spline_tools_position.x = sysinfo.desktopsize.x - 205
	if spline_tools_position.y > sysinfo.desktopsize.y - 425 do spline_tools_position.y = sysinfo.desktopsize.y - 425
	if spline_tools_position.x < 0 do spline_tools_position.x = 0 
	if spline_tools_position.y < 0 do spline_tools_position.y = 0
	createDialog spline_tools_rollout 200 365 spline_tools_position.x spline_tools_position.y
)--end local